/*
								+---------------------------------+
								|                                 |
								|   *** simple win32 window ***   |
								|                                 |
								|  Copyright   -tHE SWINe- 2006  |
								|                                 |
								|            Main.cpp             |
								|                                 |
								+---------------------------------+
*/

#include <windows.h>
#include <WinIoCtl.h>
#include <vector>
#include <algorithm>
#include "TrayIcon.h"
#include "resource.h"

static const DWORD DWORD_MAX = ((DWORD)-1 < 0)? ((DWORD)1 << (sizeof(DWORD) * 8 - 1)) |
	(((DWORD)1 << (sizeof(DWORD) * 8 - 1)) - 1) : (DWORD)-1;

class CDiskPerfMonitor {
protected:
	const char *m_p_s_drive_name;

	HANDLE m_h_device;

	DISK_PERFORMANCE m_t_diskperf;

	DWORD m_n_last_read_count;
	DWORD m_n_last_write_count;
	DWORD m_n_last_split_count;

	DWORD m_n_read_dta;
	DWORD m_n_write_dta;
	DWORD m_n_split_dta;

public:
	CDiskPerfMonitor(const char *p_s_drive_name = "\\\\.\\PhysicalDrive0")
		:m_p_s_drive_name(p_s_drive_name), m_h_device(INVALID_HANDLE_VALUE),
		m_n_last_read_count(0), m_n_last_write_count(0), m_n_last_split_count(0),
		m_n_read_dta(0), m_n_write_dta(0), m_n_split_dta(0)
	{}

	~CDiskPerfMonitor()
	{
		if(m_h_device != INVALID_HANDLE_VALUE)
			CloseHandle(m_h_device);
	}

	bool Open()
	{
		if(m_h_device != INVALID_HANDLE_VALUE)
			CloseHandle(m_h_device);

		if((m_h_device = CreateFile(m_p_s_drive_name, 0, FILE_SHARE_READ |
		   FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL)) == INVALID_HANDLE_VALUE)
			return false;

		if(!QueryPerformance())
			return false;

		m_n_read_dta = 0;
		m_n_write_dta = 0;
		m_n_split_dta = 0;
		// no deltas after first query!

		return true;
	}

	bool QueryPerformance()
	{
		m_n_read_dta = 0;
		m_n_write_dta = 0;
		m_n_split_dta = 0;
		// presume 0 on error

		if(m_h_device == INVALID_HANDLE_VALUE)
			return false;

		DWORD n_bytes_written;
		if(!DeviceIoControl(m_h_device, IOCTL_DISK_PERFORMANCE, NULL, 0,
		   &m_t_diskperf, sizeof(m_t_diskperf), &n_bytes_written, NULL))
			return false;
		// query (incremental) performance

		if(m_n_last_read_count <= m_t_diskperf.ReadCount) {
			m_n_read_dta = m_t_diskperf.ReadCount - m_n_last_read_count;
			m_n_last_read_count = m_t_diskperf.ReadCount;
		} else {
			m_n_read_dta = m_t_diskperf.ReadCount + (DWORD_MAX - m_n_last_read_count);
			m_n_last_read_count = m_t_diskperf.ReadCount;
		}
		if(m_n_last_write_count <= m_t_diskperf.WriteCount) {
			m_n_write_dta = m_t_diskperf.WriteCount - m_n_last_write_count;
			m_n_last_write_count = m_t_diskperf.WriteCount;
		} else {
			m_n_write_dta = m_t_diskperf.WriteCount + (DWORD_MAX - m_n_last_write_count);
			m_n_last_write_count = m_t_diskperf.WriteCount;
		}
		if(m_n_last_split_count <= m_t_diskperf.SplitCount) {
			m_n_split_dta = m_t_diskperf.SplitCount - m_n_last_split_count;
			m_n_last_split_count = m_t_diskperf.SplitCount;
		} else {
			m_n_split_dta = m_t_diskperf.SplitCount + (DWORD_MAX - m_n_last_split_count);
			m_n_last_split_count = m_t_diskperf.SplitCount;
		}
		// calculate delta performance

		return true;
	}

	DWORD n_DeltaOps() const
	{
		return m_n_read_dta + m_n_write_dta + m_n_split_dta;
	}

	bool b_IsActive() const
	{
		return m_n_read_dta > 0 || m_n_write_dta > 0 || m_n_split_dta > 0;
	}
};

static HWND h_wnd;
static HMENU h_popup_menu = NULL;
static HICON p_icon_single[2], p_icon_double[2][2], p_icon_triple[2][2][2];
static CTrayIcon tray;
static size_t n_mode;
static std::vector <CDiskPerfMonitor*> monitor_list;

LRESULT CALLBACK WndProc(HWND h_wnd, UINT n_msg, WPARAM n_w_param, LPARAM n_l_param);

inline void Delete(CDiskPerfMonitor *p_obj)
{
	delete p_obj;
}

void TrimSpace(std::string &r_s_string)
{
	int b = 0, e = r_s_string.length();
	while(e > 0 && isspace(r_s_string[e - 1]))
		-- e;
	while(b < e && isspace(r_s_string[b]))
		++ b;
	r_s_string.erase(e);
	r_s_string.erase(0, b);
}

bool ReadLine(FILE *p_fr, std::string &r_s_line, int &r_n_cur_line)
{
	while(!feof(p_fr)) {
		r_s_line.erase();
		try {
			for(int c = fgetc(p_fr); c != '\n' && c != EOF; c = fgetc(p_fr))
				r_s_line += c;
		} catch(std::bad_alloc&) {
			return false;
		}
		// read line

		++ r_n_cur_line;
		// line counter for file debugging

		if(r_s_line.find('#') != std::string::npos)
			r_s_line.erase(r_s_line.find('#'));
		// throw away line comment

		TrimSpace(r_s_line);
		// throw away begin / end whitespace

		if(!r_s_line.length())
			continue;
		// skip empty lines

		return true;
	}

	return false;
}

int WINAPI WinMain(HINSTANCE h_instance, HINSTANCE h_prev_instance, PSTR p_s_cmd_line, int n_cmd_show)
{
	if(!(h_wnd = CreateWindow("STATIC", "", WS_POPUP,
	   0, 0, 16, 16, NULL, NULL, h_instance, NULL)))
		return -1;
	SetWindowLong(h_wnd, GWL_WNDPROC, (LONG)&WndProc);
	ShowWindow(h_wnd, SW_HIDE);
	UpdateWindow(h_wnd);
	// create dummy window

	int n_refresh_msec;
	std::vector<std::string> device_array;
	{
		n_mode = 1;
		n_refresh_msec = 100;
		// defaults

		FILE *p_fr;
		if(!(p_fr = fopen("config.ini", "r"))) {
			MessageBox(h_wnd, "Error: Failed to open config.ini", "HDD Led", MB_OK | MB_ICONERROR);
			return -1;
		}
		int n_line = 0;
		std::string s_line;
		while(ReadLine(p_fr, s_line, n_line)) {
			int n_tmp;
			if(sscanf(s_line.c_str(), "[mode]=%d", &n_tmp) == 1)
				n_mode = min(3, max(1, n_tmp));
			else if(sscanf(s_line.c_str(), "[refresh-msec]=%d", &n_tmp) == 1)
				n_refresh_msec = max(1, n_tmp);
			else
				device_array.push_back(s_line);
		}
		fclose(p_fr);
	}
	// parse config file

	if(n_mode == 2) {
		p_icon_double[0][0] = LoadIcon(h_instance, MAKEINTRESOURCE(IDI_ICON__00));
		p_icon_double[0][1] = LoadIcon(h_instance, MAKEINTRESOURCE(IDI_ICON__01));
		p_icon_double[1][0] = LoadIcon(h_instance, MAKEINTRESOURCE(IDI_ICON__10));
		p_icon_double[1][1] = LoadIcon(h_instance, MAKEINTRESOURCE(IDI_ICON__11));
	} else if(n_mode == 3) {
		p_icon_triple[0][0][0] = LoadIcon(h_instance, MAKEINTRESOURCE(IDI_ICON_000));
		p_icon_triple[0][0][1] = LoadIcon(h_instance, MAKEINTRESOURCE(IDI_ICON_001));
		p_icon_triple[0][1][0] = LoadIcon(h_instance, MAKEINTRESOURCE(IDI_ICON_010));
		p_icon_triple[0][1][1] = LoadIcon(h_instance, MAKEINTRESOURCE(IDI_ICON_011));
		p_icon_triple[1][0][0] = LoadIcon(h_instance, MAKEINTRESOURCE(IDI_ICON_100));
		p_icon_triple[1][0][1] = LoadIcon(h_instance, MAKEINTRESOURCE(IDI_ICON_101));
		p_icon_triple[1][1][0] = LoadIcon(h_instance, MAKEINTRESOURCE(IDI_ICON_110));
		p_icon_triple[1][1][1] = LoadIcon(h_instance, MAKEINTRESOURCE(IDI_ICON_111));
	} else /*if(n_mode == 1)*/ {
		p_icon_single[0] = LoadIcon(h_instance, MAKEINTRESOURCE(IDI_ICON_OFF));
		p_icon_single[1] = LoadIcon(h_instance, MAKEINTRESOURCE(IDI_ICON_ON));
	}
	// load icons

	tray.Create(h_wnd, 0, "HDD Led", LoadIcon(NULL, IDI_APPLICATION), WM_TRAYICON_NOTIFY);
	// create tray icon

	SetTimer(h_wnd, 0, n_refresh_msec, NULL);
	// set timer

	try {
		size_t n_device_num = device_array.size();
		monitor_list.reserve(n_device_num);
		for(size_t i = 0; i < n_device_num; ++ i)
			monitor_list.push_back(new CDiskPerfMonitor(device_array[i].c_str()));
	} catch(std::bad_alloc&) {
		MessageBox(h_wnd, "Error: Not enough memory", "HDD Led", MB_OK | MB_ICONERROR);
		return -1;
	}
	// alloc monitors

	for(size_t i = 0, n = monitor_list.size(); i < n; ++ i) {
		if(!monitor_list[i]->Open()) {
			std::for_each(monitor_list.begin(), monitor_list.end(), Delete);
			MessageBox(h_wnd, "Error: Failed to open monitor for one of device(s)",
				"HDD Led", MB_OK | MB_ICONERROR);
			return -1;
		}
	}
	// open monitors

	MSG t_msg;
	while(GetMessage(&t_msg, h_wnd, 0, 0) > 0) {
		TranslateMessage(&t_msg);
		DispatchMessage(&t_msg);
	}
	// message loop

	std::for_each(monitor_list.begin(), monitor_list.end(), Delete);
	if(h_popup_menu != NULL)
		DestroyMenu(h_popup_menu);
	// clean-up

	return 0;
}

static int p_last_index[3] = {-1, -1, -1};

void On_Timer()
{
	int p_index[3] = {0};
	size_t n_max_idx = min(n_mode, sizeof(p_index) / sizeof(p_index[0])) - 1;

	for(size_t i = 0, n = monitor_list.size(); i < n; ++ i) {
		if(!monitor_list[i]->QueryPerformance())
			continue;
		if(monitor_list[i]->b_IsActive())
			p_index[min(i, n_max_idx)] = 1;
	}
	// query devices performance

	if(n_mode == 3) {
		if(p_last_index[0] != p_index[0] ||
		   p_last_index[1] != p_index[1] ||
		   p_last_index[2] != p_index[2]) {
			tray.SetIcon(p_icon_triple[p_index[0]][p_index[1]][p_index[2]]);
			p_last_index[0] = p_index[0];
			p_last_index[1] = p_index[1];
			p_last_index[2] = p_index[2];
		}
	} else if(n_mode == 2) {
		if(p_last_index[0] != p_index[0] ||
		   p_last_index[1] != p_index[1]) {
			tray.SetIcon(p_icon_double[p_index[0]][p_index[1]]);
			p_last_index[0] = p_index[0];
			p_last_index[1] = p_index[1];
		}
	} else /*if(n_mode == 1)*/ {
		if(p_last_index[0] != p_index[0]) {
			tray.SetIcon(p_icon_single[p_index[0]]);
			p_last_index[0] = p_index[0];
		}
	}
	// update tray icon (only if needed)
}

LRESULT CALLBACK WndProc(HWND h_wnd, UINT n_msg, WPARAM n_w_param, LPARAM n_l_param)
{
	switch(n_msg) {
	case WM_DESTROY:
		PostQuitMessage(0);
		return 0;
	case WM_TIMER:
		On_Timer();
		return 0;
	case WM_TRAYICON_NOTIFY:
		{
			int n_message = LOWORD(n_l_param);
			int n_icon_id = HIWORD(n_l_param); // corresponds to n_id in CTrayIcon::CTrayIcon()
			POINT t_pt;
			GetCursorPos(&t_pt);
			int n_mouse_x = t_pt.x;//LOWORD(n_w_param); // or GET_X_LPARAM
			int n_mouse_y = t_pt.y;//HIWORD(n_w_param); // or GET_Y_LPARAM
			if(n_message == WM_RBUTTONDOWN) {
				if(h_popup_menu == NULL) {
					h_popup_menu = LoadMenu(GetModuleHandle(NULL),
						MAKEINTRESOURCE(IDR_POPUP_MENU));
				}
				// load popup menu as needed

				SetForegroundWindow(h_wnd); // so the menu would disappear if clicked elsewhere
				TrackPopupMenu(GetSubMenu(h_popup_menu, 0),
					TPM_RIGHTBUTTON, n_mouse_x, n_mouse_y, 0, h_wnd, NULL);
				PostMessage(h_wnd, WM_NULL, 0, 0); // so the menu wouldn't disappear immediately next time
				// show menu
			}
		}
		break;

	case WM_COMMAND:
		switch(n_w_param) {
		case ID__EXIT:
			SendMessage(h_wnd, WM_CLOSE, 0, 0);
			break;
		case ID__ABOUT:
			MessageBox(h_wnd, "HDD Led v1.0 (built " __DATE__ ") by -tHE SWINe-",
				"HDD Led", MB_OK | MB_ICONINFORMATION);
			break;
		}
		return 0;
	}

	return DefWindowProc(h_wnd, n_msg, n_w_param, n_l_param) ;
}
